try:
    basestring
except NameError:
    basestring = str  # Python 2 -> 3 alias

""" Exposes decorator class."""

__all__ = ["DocInheritDecorator"]


class DocInheritDecorator(object):
    """ A decorator that merges provided parent docstring with the docstring of the decorated
        function/method/property.

        Methods
        -------
        doc_merger(prnt_attr_doc, child_doc)
            Merges the parent and child docstrings into a single docstring.

        Notes
        -----
        When utilized as the inner-most decorator, this decorator can be used on functions decorated functions,
        methods, properties, static methods, class methods, and their abstract counterparts."""

    def __init__(self, prnt_doc):
        """ Parameters
            ----------
            prnt_doc : Union[str, Any]
                The docstring, or object of which the docstring is utilized as the
                parent docstring during the docstring merge."""
        self.prnt_doc = prnt_doc if isinstance(prnt_doc, basestring) else prnt_doc.__doc__

    def __call__(self, func):
        """ Parameters
            ----------
            func : FunctionType
                The function/method/property to be decorated

            Returns
            -------
            FunctionType
                The decorated function/method/property whose docstring is given by
                DocInheritDecorator.doc_merger(prnt_attr_doc, child_doc)"""
        func.__doc__ = self.doc_merger(self.prnt_doc, func.__doc__)
        return func

    @staticmethod
    def doc_merger(prnt_attr_doc, child_doc):
        """ Merges the parent and child docstrings into a single docstring.

            Parameters
            ----------
            prnt_cls_doc: Union[None, str]
            child_doc: Union[None, str]

            Raises
            ------
            NotImplementedError"""
        raise NotImplementedError



from __future__ import absolute_import
from ._doc_parse_tools import merge_numpy_docs, merge_rest_docs, merge_google_napoleon_docs, merge_numpy_napoleon_docs

""" Docstring inheritance-style implementations.

    To implement your own inheritance file, simply write a function that fits the template:

            def your_style(prnt_doc, child_doc):
                ''' Merges parent and child docstrings

                    Parameters
                    ----------
                    prnt_cls_doc: Optional[str]
                    child_doc: Optional[str]

                    Returns
                    ------
                    Optional[str]
                        The merged docstring that will be utilized.'''
                return final_docstring

    and log this using `custom_inherit.add_style(your_style)`. To permanently save your function,
    define your function within custom_inherit/_style_store.py, and log it in custom_inherit.style_store.__all__.
    Your style will then be available as 'your_style' (i.e. whatever you named the function).
"""

# All built-in styles must be logged in the __all__ field.
__all__ = ["parent", "numpy", "reST", "google", "numpy_napoleon"]


def parent(prnt_doc, child_doc):
    """ Wherever the docstring for a child-class' attribute (or for the class itself) is
        `None`, inherit the corresponding docstring from the parent.

        *NOTE* As of Python 3.5, this is the default behavior of the built-in function
        inspect.getdoc, and thus this style is deprecated Python 3.(>=5). """
    return child_doc if child_doc is not None else prnt_doc


def numpy(prnt_doc, child_doc):
    """ Merges numpy-styled docstrings from the parent and child.

        Specifically, any docstring section that appears in the parent's docstring that
        is not present in the child's is inherited. Otherwise, the child's docstring
        section is utilized. An exception to this is if the parent docstring contains a
        "Raises" section, but the child's attribute's docstring contains a "Returns" or
        "Yields" section instead. In this instance, the "Raises" section will not appear in the
        inherited docstring.

        Any whitespace that can be uniformly removed from a docstring's second line and onwards is
        removed. Sections in the resulting docstring will be separated by a single blank line.

        For details on the numpy docstring style, see:
        https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt

        Example:
            - parent's docstring:

                ''' Parent's line

                    Parameters
                    ----------
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                    Raises
                    ------
                    NotImplemented Error'''

            - child's docstring:

                ''' Child's line

                    Returns
                    -------
                    int

                    Notes
                    -----
                    notes blah blah'''

            - docstring that is ultimately inherited:

                ''' Child's line

                    Parameters
                    ----------
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                    Returns
                    -------
                    int

                    Notes
                    -----
                    notes blah blah'''"""
    return merge_numpy_docs(prnt_doc, child_doc)


def reST(prnt_doc, child_doc):
    """ Merge two reST-style docstrings into a single docstring.

        Given the reST-style docstrings from a parent and child's attributes, merge the docstring
        sections such that the child's section is used, wherever present, otherwise the parent's
        section is used.

        Sections are delimited by any type of reST section title. For more details, see:
        http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#sections

        Any whitespace that can be uniformly removed from a docstring's second line and onwards is
        removed. Sections in the resulting docstring will be separated by a single blank line.

        Example:
          parent - ''' Header1
                       -------
                       parent's content for Header 1
                           - indented material
                       Header2
                       ~~~~~~~
                       content for Header2'''

          child -  ''' Front-matter
                       ~~~~~~~~~
                       NewHeader
                       ~~~~~~~~~
                       content for NewHeader

                       Header2
                       +++++++
                       child's content for Header2'''

          merged - ''' Front-matter

                       Header1
                       -------
                       content for Header 1
                           - indented material

                       Header2
                       +++++++
                       child's content for Header2

                       ~~~~~~~~~
                       NewHeader
                       ~~~~~~~~~
                       content for NewHeader '''
            """
    return merge_rest_docs(prnt_doc, child_doc)


def numpy_napoleon(prnt_doc, child_doc):
    """ Behaves identically to the 'numpy' style, but abides by the docstring sections
        specified by the "Napoleon" standard.

        For more info regarding the Napoleon standard, see:
        http://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html#docstring-sections

        Example:
            - parent's docstring:

                ''' Parent's line

                    Keyword Arguments
                    -----------------
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                    Raises
                    ------
                    NotImplemented Error'''

            - child's docstring:

                ''' Child's line

                    Returns
                    -------
                    int

                    Notes
                    -----
                    notes blah blah'''

            - docstring that is ultimately inherited:

                ''' Child's line

                    Keyword Arguments
                    -----------------
                    x: int
                        description of x
                    y: Union[None, int]
                        description of y

                    Returns
                    -------
                    int

                    Notes
                    -----
                    notes blah blah'''
    """
    return merge_numpy_napoleon_docs(prnt_doc, child_doc)


def google(prnt_doc, child_doc):
    """ Merges google-styled docstrings from the parent and child, abiding to the docstring sections
        specified by the "Napoleon" standard.

        Specifically, any docstring section that appears in the parent's docstring that
        is not present in the child's is inherited. Otherwise, the child's docstring
        section is utilized. An exception to this is if the parent docstring contains a
        "Raises" section, but the child's attribute's docstring contains a "Returns" or
        "Yields" section instead. In this instance, the "Raises" section will not appear in the
        inherited docstring.

        Any whitespace that can be uniformly removed from a docstring's second line and onwards is
        removed. Sections in the resulting docstring will be separated by a single blank line.

        For more info regarding the Napoleon standard, see:
        http://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html#docstring-sections

        For details on the google docstring style, see:
        http://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html#example-google

        Example:
            - parent's docstring:

                ''' Parent's line

                    Args:
                        x: int
                            description of x
                        y: Union[None, int]
                            description of y

                    Raises:
                        NotImplemented Error'''

            - child's docstring:

                ''' Child's line

                    Returns:
                        int

                    Notes:
                        notes blah blah'''

            - docstring that is ultimately inherited:

                ''' Child's line

                    Parameters:
                        x: int
                            description of x
                        y: Union[None, int]
                            description of y

                    Returns:
                        int

                    Notes:
                        notes blah blah'''
        """
    return merge_google_napoleon_docs(prnt_doc, child_doc)



from __future__ import absolute_import

from .numpy_parse_tools import merge_numpy_docs
from .rest_parse_tools import merge_rest_docs
from .napoleon_parse_tools import merge_numpy_napoleon_docs, merge_google_napoleon_docs

__all__ = ["merge_numpy_docs", "merge_rest_docs", "merge_numpy_napoleon_docs", "merge_google_napoleon_docs"]



from __future__ import absolute_import
from collections import OrderedDict
from inspect import cleandoc
from string import punctuation

__all__ = ["merge_rest_docs"]


def is_delimiter(line):
    """ True if a line consists only of a single punctuation character."""
    return bool(line) and line[0] in punctuation and line[0]*len(line) == line


def parse_rest_doc(doc):
    """ Extract the headers, delimiters, and text from reST-formatted docstrings.

        Parameters
        ----------
        doc: Union[str, None]

        Returns
        -------
        Dict[str, Section] """

    class Section(object):
        def __init__(self, header=None, body=None):
            self.header = header  # str
            self.body = body  # str

    doc_sections = OrderedDict([('', Section(header=''))])
    if not doc:
        return doc_sections

    doc = cleandoc(doc)
    lines = iter(doc.splitlines())

    header = ''
    body = []
    section = Section(header=header)
    line = ''
    while True:
        try:
            prev_line = line
            line = next(lines)
            # section header encountered
            if is_delimiter(line) and 0 < len(prev_line) <= len(line):
                # prev-prev-line is overline
                if len(body) >= 2 and len(body[-2]) == len(line) \
                        and body[-2][0] == line[0] and is_delimiter(body[-2]):
                    lim = -2
                else:
                    lim = -1

                section.body = "\n".join(body[:lim]).rstrip()
                doc_sections.update([(header.strip(), section)])
                section = Section(header="\n".join(body[lim:] + [line]))
                header = prev_line
                body = []
                line = ''
            else:
                body.append(line)

        except StopIteration:
            section.body = "\n".join(body).rstrip()
            doc_sections.update([(header.strip(), section)])
            break
    return doc_sections


def merge_rest_docs(prnt_doc=None, child_doc=None):
    """ See custom_inherit.style_store.reST for details. """
    prnt_sections = parse_rest_doc(prnt_doc)
    child_sections = parse_rest_doc(child_doc)

    header = prnt_sections['']
    prnt_sections.update(child_sections)
    if not child_sections[''].body:
        prnt_sections[''] = header
        if not header.body:
            prnt_sections.popitem(last=False)

    return "\n\n".join(("\n".join((x.header, x.body)) for x in prnt_sections.values())).lstrip()
